# Copyright 2010 OpenStack Foundation
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_service import wsgi
from oslo_utils import uuidutils

import routes
import webob
import webob.dec
import webob.request

from karbor.api.openstack import wsgi as os_wsgi
from karbor import context
from karbor.services.protection.protection_plugins.volume \
    import volume_plugin_cinder_schemas as cinder_schemas

FAKE_UUID = 'aaaaaaaa-aaaa-aaaa-aaaa-aaaaaaaaaaaa'
FAKE_UUIDS = {}
PROVIDER_OS = {
    "description": "This provider uses OpenStack's own services "
                   "(swift, cinder) as storage",
    "extended_info_schema": {
        "options_schema": {
            "OS::Cinder::Volume": cinder_schemas.OPTIONS_SCHEMA
        },
        "saved_info_schema": {
            "OS::Cinder::Volume": cinder_schemas.SAVED_INFO_SCHEMA
        },
        "restore_schema": {
            "OS::Cinder::Volume": cinder_schemas.RESTORE_SCHEMA
        }
    },
    "id": "cf56bd3e-97a7-4078-b6d5-f36246333fd9",
    "name": "OS Infra Provider"
}


class Context(object):
    pass


class FakeRouter(wsgi.Router):
    def __init__(self, ext_mgr=None):
        pass

    @webob.dec.wsgify
    def __call__(self, req):
        res = webob.Response()
        res.status = '200'
        res.headers['X-Test-Success'] = 'True'
        return res


@webob.dec.wsgify
def fake_wsgi(self, req):
    return self.application


class FakeToken(object):
    id_count = 0

    def __getitem__(self, key):
        return getattr(self, key)

    def __init__(self, **kwargs):
        FakeToken.id_count += 1
        self.id = FakeToken.id_count
        for k, v in kwargs.items():
            setattr(self, k, v)


class FakeRequestContext(context.RequestContext):
    def __init__(self, *args, **kwargs):
        kwargs['auth_token'] = kwargs.get('auth_token', 'fake_auth_token')
        super(FakeRequestContext, self).__init__(*args, **kwargs)


class HTTPRequest(webob.Request):

    @classmethod
    def blank(cls, *args, **kwargs):
        if args is not None:
            if args[0].find('v1') == 0:
                kwargs['base_url'] = 'http://localhost/v1'
            else:
                kwargs['base_url'] = 'http://localhost/v2'

        use_admin_context = kwargs.pop('use_admin_context', False)
        out = os_wsgi.Request.blank(*args, **kwargs)
        out.environ['karbor.context'] = FakeRequestContext(
            'fake_user',
            'fakeproject',
            is_admin=use_admin_context)
        return out


class TestRouter(wsgi.Router):
    def __init__(self, controller):
        mapper = routes.Mapper()
        mapper.resource("test", "tests",
                        controller=os_wsgi.Resource(controller))
        super(TestRouter, self).__init__(mapper)


def get_fake_uuid(token=0):
    if token not in FAKE_UUIDS:
        FAKE_UUIDS[token] = uuidutils.generate_uuid()
    return FAKE_UUIDS[token]



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from datetime import timedelta
import time

from karbor.common import constants
from karbor import context
from karbor import objects
from karbor.services.operationengine.engine.executors import \
    thread_pool_executor
from karbor.tests import base


class FakeOperationManager(object):
    def run_operation(self, operation_type, operation_definition, **kwargs):
        return


class ThreadPoolExecutorTestCase(base.TestCase):

    def setUp(self):
        super(ThreadPoolExecutorTestCase, self).setUp()

        self._operation_manager = FakeOperationManager()
        self._executor = thread_pool_executor.ThreadPoolExecutor(
            self._operation_manager)
        self.context = context.get_admin_context()

    def tearDown(self):
        super(ThreadPoolExecutorTestCase, self).tearDown()
        self._executor.shutdown()

    def test_execute_operation(self):
        operation = self._create_operation()
        self._create_operation_state(operation.id, 0)

        now = datetime.utcnow()
        window_time = 30

        self._executor.execute_operation(operation.id, now, now, window_time)

        time.sleep(1)

        self.assertEqual(0, len(self._executor._operation_to_run))

        state = objects.ScheduledOperationState.get_by_operation_id(
            self.context, operation.id)
        self.assertIsNotNone(state.end_time_for_run)
        self.assertEqual(constants.OPERATION_STATE_REGISTERED, state.state)

    def test_resume_operation(self):
        operation = self._create_operation()
        self._create_operation_state(operation.id, 0)

        now = datetime.utcnow()
        window_time = 30

        self._executor.resume_operation(operation.id, end_time_for_run=(
            now + timedelta(seconds=window_time)))

        time.sleep(1)

        self.assertEqual(0, len(self._executor._operation_to_run))

        state = objects.ScheduledOperationState.get_by_operation_id(
            self.context, operation.id)
        self.assertEqual(constants.OPERATION_STATE_REGISTERED, state.state)

    def test_cancel_operation(self):
        operation_id = '123'

        self._executor.cancel_operation(operation_id)
        self.assertEqual(0, len(self._executor._operation_to_cancel))

        self._executor._operation_to_run[operation_id] = 0
        self._executor.cancel_operation(operation_id)
        self.assertEqual(1, len(self._executor._operation_to_cancel))

    def _create_operation(self, trigger_id='123'):
        operation_info = {
            'name': 'protect vm',
            'description': 'protect vm resource',
            'operation_type': 'protect',
            'user_id': '123',
            'project_id': '123',
            'trigger_id': trigger_id,
            'operation_definition': {}
        }
        operation = objects.ScheduledOperation(self.context, **operation_info)
        operation.create()
        return operation

    def _create_operation_state(self, operation_id, service_id):
        state_info = {
            'operation_id': operation_id,
            'service_id': service_id,
            'trust_id': '123',
            'state': constants.OPERATION_STATE_INIT,
        }
        state = objects.ScheduledOperationState(self.context, **state_info)
        state.create()
        return state



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_versionedobjects import fields

from karbor import objects


def fake_db_restore(**updates):
    db_restore = {
        "id": "36ea41b2-c358-48a7-9117-70cb7617410a",
        "project_id": "586cc6ce-e286-40bd-b2b5-dd32694d9944",
        "provider_id": "2220f8b1-975d-4621-a872-fa9afb43cb6c",
        "checkpoint_id": "09edcbdc-d1c2-49c1-a212-122627b20968",
        "restore_target": "192.168.1.2/identity/",
        "parameters": '{}',
        "restore_auth": '{"type": "password", "username": "admin",'
                        '"password": "test" }',
        "status": "SUCCESS"
    }
    for name, field in objects.Restore.fields.items():
        if name in db_restore:
            continue
        if field.nullable:
            db_restore[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_restore[name] = field.default
        else:
            raise Exception('fake_db_restore needs help with %s.' % name)

    if updates:
        db_restore.update(updates)

    return db_restore



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Handles all requests relating to protection service."""


from karbor.db import base
from karbor.services.protection import rpcapi as protection_rpcapi


class API(base.Base):
    """API for interacting with the protection manager."""

    def __init__(self, db_driver=None):
        self.protection_rpcapi = protection_rpcapi.ProtectionAPI()
        super(API, self).__init__(db_driver)

    def restore(self, context, restore, restore_auth):
        return self.protection_rpcapi.restore(context, restore, restore_auth)

    def protect(self, context, plan, checkpoint_properties):
        return self.protection_rpcapi.protect(context, plan,
                                              checkpoint_properties)

    def delete(self, context, provider_id, checkpoint_id):
        return self.protection_rpcapi.delete(
            context,
            provider_id,
            checkpoint_id
        )

    def show_checkpoint(self, context, provider_id, checkpoint_id):
        return self.protection_rpcapi.show_checkpoint(
            context,
            provider_id,
            checkpoint_id
        )

    def list_checkpoints(self, context, provider_id, marker, limit,
                         sort_keys, sort_dirs, filters, offset):
        return self.protection_rpcapi.list_checkpoints(
            context,
            provider_id,
            marker,
            limit,
            sort_keys,
            sort_dirs,
            filters
        )

    def list_protectable_types(self, context):
        return self.protection_rpcapi.list_protectable_types(context)

    def show_protectable_type(self, context, protectable_type):
        return self.protection_rpcapi.show_protectable_type(
            context,
            protectable_type
        )

    def list_protectable_instances(self, context, protectable_type,
                                   marker, limit, sort_keys,
                                   sort_dirs, filters, offset, parameters):
        return self.protection_rpcapi.list_protectable_instances(
            context,
            protectable_type,
            marker,
            limit,
            sort_keys,
            sort_dirs,
            filters,
            parameters
        )

    def list_protectable_dependents(self, context,
                                    protectable_id,
                                    protectable_type):
        return self.protection_rpcapi.list_protectable_dependents(
            context,
            protectable_id,
            protectable_type
        )

    def show_protectable_instance(self, context,
                                  protectable_type,
                                  protectable_id,
                                  parameters=None):
        return self.protection_rpcapi.show_protectable_instance(
            context,
            protectable_type,
            protectable_id,
            parameters=parameters
        )

    def show_provider(self, context, provider_id):
        return self.protection_rpcapi.show_provider(context, provider_id)

    def list_providers(self, context, marker, limit,
                       sort_keys, sort_dirs, filters, offset):
        return self.protection_rpcapi.list_providers(
            context,
            marker,
            limit,
            sort_keys,
            sort_dirs,
            filters
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
from datetime import timedelta
import eventlet
import mock
from oslo_config import cfg

from karbor import exception
from karbor.services.operationengine.engine.triggers.timetrigger.time_trigger \
    import TimeTrigger
from karbor.tests import base


class FakeTimeFormat(object):
    def __init__(self, start_time, pattern):
        super(FakeTimeFormat, self).__init__()

    @classmethod
    def check_time_format(cls, pattern):
        pass

    def compute_next_time(self, current_time):
        return current_time + timedelta(seconds=0.5)

    def get_min_interval(self):
        return cfg.CONF.min_interval


class FakeExecutor(object):
    def __init__(self):
        super(FakeExecutor, self).__init__()
        self._ops = {}

    def execute_operation(self, operation_id, triggered_time,
                          expect_start_time, window):
        if operation_id not in self._ops:
            self._ops[operation_id] = 0
        self._ops[operation_id] += 1
        eventlet.sleep(0.5)

    def clear(self):
        self._ops.clear()


class TimeTriggerTestCase(base.TestCase):

    def setUp(self):
        super(TimeTriggerTestCase, self).setUp()

        self._set_configuration()

        mock_obj = mock.Mock()
        mock_obj.return_value = FakeTimeFormat
        TimeTrigger._get_time_format_class = mock_obj

        self._default_executor = FakeExecutor()

    def test_check_configuration(self):
        self._set_configuration(10, 20, 30)
        self.assertRaisesRegex(exception.InvalidInput,
                               "Configurations of time trigger are invalid",
                               TimeTrigger.check_configuration)
        self._set_configuration()

    def test_check_trigger_property_start_time(self):
        trigger_property = {
            "pattern": "",
            "start_time": ""
        }

        self.assertRaisesRegex(exception.InvalidInput,
                               "The trigger\'s start time is unknown",
                               TimeTrigger.check_trigger_definition,
                               trigger_property)

        trigger_property['start_time'] = 'abc'
        self.assertRaisesRegex(exception.InvalidInput,
                               "The format of trigger .* is not correct",
                               TimeTrigger.check_trigger_definition,
                               trigger_property)

        trigger_property['start_time'] = 123
        self.assertRaisesRegex(exception.InvalidInput,
                               "The trigger .* is not an instance of string",
                               TimeTrigger.check_trigger_definition,
                               trigger_property)

    @mock.patch.object(FakeTimeFormat, 'get_min_interval')
    def test_check_trigger_property_interval(self, get_min_interval):
        get_min_interval.return_value = 0

        trigger_property = {
            "start_time": '2016-8-18 01:03:04'
        }

        self.assertRaisesRegex(exception.InvalidInput,
                               "The interval of two adjacent time points .*",
                               TimeTrigger.check_trigger_definition,
                               trigger_property)

    def test_check_trigger_property_window(self):
        trigger_property = {
            "window": "abc",
            "start_time": '2016-8-18 01:03:04'
        }

        self.assertRaisesRegex(exception.InvalidInput,
                               "The trigger window.* is not integer",
                               TimeTrigger.check_trigger_definition,
                               trigger_property)

        trigger_property['window'] = 1000
        self.assertRaisesRegex(exception.InvalidInput,
                               "The trigger windows .* must be between .*",
                               TimeTrigger.check_trigger_definition,
                               trigger_property)

    def test_check_trigger_property_end_time(self):
        trigger_property = {
            "window": 15,
            "start_time": '2016-8-18 01:03:04',
            "end_time": "abc"
        }

        self.assertRaisesRegex(exception.InvalidInput,
                               "The format of trigger .* is not correct",
                               TimeTrigger.check_trigger_definition,
                               trigger_property)

    def test_register_operation(self):
        trigger = self._generate_trigger()

        operation_id = "1"
        trigger.register_operation(operation_id)
        eventlet.sleep(0.3)

        self.assertGreaterEqual(trigger._executor._ops[operation_id], 1)
        self.assertRaisesRegex(exception.ScheduledOperationExist,
                               "The operation_id.* is exist",
                               trigger.register_operation,
                               operation_id)

        eventlet.sleep(0.3)
        self.assertRaises(exception.TriggerIsInvalid,
                          trigger.register_operation,
                          "2")

    def test_unregister_operation(self):
        trigger = self._generate_trigger()
        operation_id = "2"

        trigger.register_operation(operation_id)
        self.assertIn(operation_id, trigger._operation_ids)

        trigger.unregister_operation(operation_id)
        self.assertNotIn(operation_id, trigger._operation_ids)

    def test_unregister_operation_when_scheduling(self):
        trigger = self._generate_trigger()

        for op_id in ['1', '2', '3']:
            trigger.register_operation(op_id)
            self.assertIn(op_id, trigger._operation_ids)
        eventlet.sleep(0.5)

        for op_id in ['2', '3']:
            trigger.unregister_operation(op_id)
            self.assertNotIn(op_id, trigger._operation_ids)
        eventlet.sleep(0.6)

        self.assertGreaterEqual(trigger._executor._ops['1'], 1)

        self.assertTrue(('2' not in trigger._executor._ops) or (
            '3' not in trigger._executor._ops))

    def test_update_trigger_property(self):
        trigger = self._generate_trigger()

        trigger_property = {
            "pattern": "",
            "window": 15,
            "start_time": datetime.utcnow(),
            "end_time": datetime.utcnow()
        }

        self.assertRaisesRegex(exception.InvalidInput,
                               ".*Can not find the first run tim",
                               trigger.update_trigger_property,
                               trigger_property)

        trigger.register_operation('1')
        eventlet.sleep(0.2)
        trigger_property['end_time'] = (
            datetime.utcnow() + timedelta(seconds=1))
        self.assertRaisesRegex(exception.InvalidInput,
                               ".*First run time.* must be after.*",
                               trigger.update_trigger_property,
                               trigger_property)

    def test_update_trigger_property_success(self):
        trigger = self._generate_trigger()
        trigger.register_operation('1')
        eventlet.sleep(0.2)

        trigger_property = {
            "pattern": "",
            "window": 15,
            "start_time": datetime.utcnow(),
            "end_time": ''
        }
        with mock.patch.object(FakeTimeFormat, 'compute_next_time') as c:
            c.return_value = datetime.utcnow() + timedelta(seconds=20)
            old_id = id(trigger._greenthread)

            trigger.update_trigger_property(trigger_property)

            self.assertNotEqual(old_id, id(trigger._greenthread))

    def _generate_trigger(self, end_time=None):
        if not end_time:
            end_time = datetime.utcnow() + timedelta(seconds=1)

        trigger_property = {
            "pattern": "",
            "window": 15,
            "start_time": datetime.utcnow(),
            "end_time": end_time
        }

        self._default_executor.clear()
        return TimeTrigger("123", trigger_property, self._default_executor)

    def _set_configuration(self, min_window=15,
                           max_window=30, min_interval=60):
        self.override_config('min_interval', min_interval)
        self.override_config('min_window_time', min_window)
        self.override_config('max_window_time', max_window)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from karbor import exception
from karbor.exception import ListProtectableResourceFailed
from karbor.i18n import _
from karbor.services.protection.graph import build_graph
import six

from oslo_log import log as logging
from stevedore import extension

LOG = logging.getLogger(__name__)


class ProtectablePluginLoadFailed(exception.KarborException):
    message = _("Could not load %(name)s: %(error)s")


def _raise_extension_exception(extmanager, ep, err):
    raise ProtectablePluginLoadFailed(name=ep.name,
                                      error=six.text_type(err))


class ProtectableRegistry(object):

    def __init__(self):
        super(ProtectableRegistry, self).__init__()
        self._protectable_map = {}
        self._plugin_map = {}

    def load_plugins(self):
        """Load all protectable plugins configured and register them.

        """
        mgr = extension.ExtensionManager(
            namespace='karbor.protectables',
            invoke_on_load=True,
            on_load_failure_callback=_raise_extension_exception)

        for e in mgr:
            self.register_plugin(e.obj)

    def register_plugin(self, plugin):
        self._plugin_map[plugin.get_resource_type()] = plugin

    def _get_protectable(self, context, resource_type):
        if resource_type in self._protectable_map:
            return self._protectable_map[resource_type]

        protectable = self._plugin_map[resource_type].instance(context)
        self._protectable_map[resource_type] = protectable
        return protectable

    def list_resource_types(self):
        """List all resource types supported by protectables.

        :return: The list of supported resource types.
        """
        return [type for type in self._plugin_map.keys()]

    def get_protectable_resource_plugin(self, resource_type):
        """Get the protectable plugin with the specified type."""
        return self._plugin_map.get(resource_type)

    def list_resources(self, context, resource_type, parameters=None):
        """List resource instances of given type.

        :param resource_type: The resource type to list instance.
        :return: The list of resource instance.
        """
        protectable = self._get_protectable(context, resource_type)
        return protectable.list_resources(context, parameters=parameters)

    def show_resource(self, context, resource_type, resource_id,
                      parameters=None):
        """List resource instances of given type.

        :param resource_type: The resource type of instance.
        :param resource_id: The resource id of instance.
        :return: The show of resource instance.
        """
        protectable = self._get_protectable(context, resource_type)
        return protectable.show_resource(context, resource_id,
                                         parameters=parameters)

    def fetch_dependent_resources(self, context, resource):
        """List dependent resources under given parent resource.

        :param resource: The parent resource to list dependent resources.
        :return: The list of dependent resources.
        """
        result = []
        for plugin in self._plugin_map.values():
            if resource.type in plugin.get_parent_resource_types():
                protectable = self._get_protectable(
                    context,
                    plugin.get_resource_type())
                try:
                    protectable_resources = \
                        protectable.get_dependent_resources(context, resource)
                except ListProtectableResourceFailed as e:
                    LOG.error("List resources failed, so skip it. "
                              "Error: {0}".format(e))
                    protectable_resources = []
                result.extend(protectable_resources)

        return result

    def build_graph(self, context, resources):
        def fetch_dependent_resources_context(resource):
            return self.fetch_dependent_resources(context, resource)

        return build_graph(
            start_nodes=resources,
            get_child_nodes_func=fetch_dependent_resources_context,
        )



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import os

from karbor import exception
from karbor.i18n import _
from karbor.services.protection import bank_plugin
from karbor.services.protection.checkpoint import CheckpointCollection
from karbor import utils
from oslo_config import cfg
from oslo_log import log as logging

provider_opts = [
    cfg.MultiStrOpt('plugin',
                    default='',
                    help='plugins to use for protection'),
    cfg.StrOpt('bank',
               default='',
               help='bank plugin to use for storage'),
    cfg.StrOpt('description',
               default='',
               help='the description of provider'),
    cfg.StrOpt('name',
               default='',
               help='the name of provider'),
    cfg.StrOpt('id',
               default='',
               help='the provider id'),
    cfg.BoolOpt('enabled',
                default=False,
                help='enabled or not'),
]
CONF = cfg.CONF

LOG = logging.getLogger(__name__)

PROTECTION_NAMESPACE = 'karbor.protections'

CONF.register_opt(cfg.StrOpt('provider_config_dir',
                             default='providers.d',
                             help='Configuration directory for providers.'
                                  ' Absolute path, or relative to karbor '
                                  ' configuration directory.'))


class PluggableProtectionProvider(object):
    def __init__(self, provider_config):
        super(PluggableProtectionProvider, self).__init__()
        self._config = provider_config
        self._id = self._config.provider.id
        self._name = self._config.provider.name
        self._description = self._config.provider.description
        self._extended_info_schema = {'options_schema': {},
                                      'restore_schema': {},
                                      'saved_info_schema': {}}
        self.checkpoint_collection = None
        self._bank_plugin = None
        self._plugin_map = {}

        if (hasattr(self._config.provider, 'bank') and
                not self._config.provider.bank):
            raise ImportError(_("Empty bank"))

        self._load_bank(self._config.provider.bank)
        self._bank = bank_plugin.Bank(self._bank_plugin)
        self.checkpoint_collection = CheckpointCollection(
            self._bank)

        if hasattr(self._config.provider, 'plugin'):
            for plugin_name in self._config.provider.plugin:
                if not plugin_name:
                    raise ImportError(_("Empty protection plugin"))
                self._register_plugin(plugin_name)

    @property
    def id(self):
        return self._id

    @property
    def name(self):
        return self._name

    @property
    def description(self):
        return self._description

    @property
    def extended_info_schema(self):
        return self._extended_info_schema

    @property
    def bank(self):
        return self._bank

    @property
    def plugins(self):
        return self._plugin_map

    def load_plugins(self):
        return {
            plugin_type: plugin_class(self._config)
            for plugin_type, plugin_class in self.plugins.items()
        }

    def _load_bank(self, bank_name):
        try:
            plugin = utils.load_plugin(PROTECTION_NAMESPACE, bank_name,
                                       self._config)
        except Exception:
            LOG.exception("Load bank plugin: '%s' failed.", bank_name)
            raise
        else:
            self._bank_plugin = plugin

    def _register_plugin(self, plugin_name):
        try:
            plugin = utils.load_class(PROTECTION_NAMESPACE, plugin_name)
        except Exception:
            LOG.exception("Load protection plugin: '%s' failed.", plugin_name)
            raise
        else:
            for resource in plugin.get_supported_resources_types():
                self._plugin_map[resource] = plugin
                if hasattr(plugin, 'get_options_schema'):
                    self._extended_info_schema['options_schema'][resource] \
                        = plugin.get_options_schema(resource)
                if hasattr(plugin, 'get_restore_schema'):
                    self._extended_info_schema['restore_schema'][resource] \
                        = plugin.get_restore_schema(resource)
                if hasattr(plugin, 'get_saved_info_schema'):
                    self._extended_info_schema['saved_info_schema'][resource] \
                        = plugin.get_saved_info_schema(resource)

    def get_checkpoint_collection(self):
        return self.checkpoint_collection

    def get_checkpoint(self, checkpoint_id):
        return self.get_checkpoint_collection().get(checkpoint_id)

    def list_checkpoints(self, provider_id, limit=None, marker=None,
                         plan_id=None, start_date=None, end_date=None,
                         sort_dir=None):
        checkpoint_collection = self.get_checkpoint_collection()
        return checkpoint_collection.list_ids(
            provider_id=provider_id, limit=limit, marker=marker,
            plan_id=plan_id, start_date=start_date, end_date=end_date,
            sort_dir=sort_dir)


class ProviderRegistry(object):
    def __init__(self):
        super(ProviderRegistry, self).__init__()
        self.providers = {}
        self._load_providers()

    def _load_providers(self):
        """load provider"""
        config_dir = utils.find_config(CONF.provider_config_dir)

        for config_file in os.listdir(config_dir):
            if not config_file.endswith('.conf'):
                continue
            config_path = os.path.abspath(os.path.join(config_dir,
                                                       config_file))
            provider_config = cfg.ConfigOpts()
            provider_config(args=['--config-file=' + config_path])
            provider_config.register_opts(provider_opts, 'provider')

            provider_enabled = provider_config.provider.enabled
            if not provider_enabled:
                LOG.info('Provider {0} is not enabled'.format(
                    provider_config.provider.name)
                )
                continue

            try:
                provider = PluggableProtectionProvider(provider_config)
            except Exception as e:
                LOG.error("Load provider: %(provider)s failed. "
                          "Reason: %(reason)s",
                          {'provider': provider_config.provider.name,
                           'reason': e})
            else:
                LOG.info('Loaded provider: %s successfully.',
                         provider_config.provider.name)
                self.providers[provider.id] = provider

    def list_providers(self, marker=None, limit=None, sort_keys=None,
                       sort_dirs=None, filters=None):
        # TODO(wangliuan) How to use the list option
        return [dict(id=provider.id, name=provider.name,
                     description=provider.description,
                     extended_info_schema=provider.extended_info_schema)
                for provider in self.providers.values()]

    def show_provider(self, provider_id):
        try:
            return self.providers[provider_id]
        except KeyError:
            raise exception.ProviderNotFound(provider_id=provider_id)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_serialization import jsonutils
from oslo_versionedobjects import fields
import six

from karbor import db
from karbor import exception
from karbor.i18n import _
from karbor import objects
from karbor.objects import base


@base.KarborObjectRegistry.register
class Restore(base.KarborPersistentObject, base.KarborObject,
              base.KarborObjectDictCompat,
              base.KarborComparableObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'id': fields.UUIDField(),
        'project_id': fields.UUIDField(),
        'provider_id': fields.UUIDField(),
        'checkpoint_id': fields.UUIDField(),
        'restore_target': fields.StringField(nullable=True),
        'parameters': base.DictOfDictOfStringsField(nullable=True),
        'status': fields.StringField(nullable=True),
        'resources_status': fields.DictOfStringsField(nullable=True),
        'resources_reason': fields.DictOfStringsField(nullable=True),
    }

    json_fields = ('parameters', 'resources_status', 'resources_reason')

    @classmethod
    def _from_db_object(cls, context, restore, db_restore):
        for name, field in restore.fields.items():
            value = db_restore.get(name)
            if isinstance(field, fields.IntegerField):
                value = value or 0
            elif isinstance(field, fields.DateTimeField):
                value = value or None
            if name in cls.json_fields:
                value = jsonutils.loads(value) if value else {}
            restore[name] = value

        restore._context = context
        restore.obj_reset_changes()
        return restore

    @classmethod
    def _convert_properties_to_db_format(cls, updates):
        for attr in cls.json_fields:
            value = updates.pop(attr, None)
            if value:
                updates[attr] = jsonutils.dumps(value)

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason=_('already created'))
        updates = self.karbor_obj_get_changes()
        self._convert_properties_to_db_format(updates)
        db_restore = db.restore_create(self._context, updates)
        self._from_db_object(self._context, self, db_restore)

    @base.remotable
    def save(self):
        updates = self.karbor_obj_get_changes()
        self._convert_properties_to_db_format(updates)
        if updates:
            db.restore_update(self._context, self.id, updates)
            self.obj_reset_changes()

    @base.remotable
    def destroy(self):
        with self.obj_as_admin():
            db.restore_destroy(self._context, self.id)

    @base.remotable
    def update_resource_status(self, resource_type, resource_id, status,
                               reason=None):
        key = '{}#{}'.format(resource_type, resource_id)
        if not self.obj_attr_is_set('resources_status'):
            self.resources_status = {}
        self.resources_status[key] = status
        self._changed_fields.add('resources_status')
        if isinstance(reason, six.string_types):
            if not self.obj_attr_is_set('resources_reason'):
                self.resources_reason = {}
            self.resources_reason[key] = reason
            self._changed_fields.add('resources_reason')


@base.KarborObjectRegistry.register
class RestoreList(base.ObjectListBase, base.KarborObject):
    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('Restore'),
    }

    @base.remotable_classmethod
    def get_all(cls, context, marker, limit, sort_keys=None, sort_dirs=None,
                filters=None, offset=None):
        restores = db.restore_get_all(context, marker, limit,
                                      sort_keys=sort_keys, sort_dirs=sort_dirs,
                                      filters=filters, offset=offset)
        return base.obj_make_list(context, cls(context), objects.Restore,
                                  restores)

    @base.remotable_classmethod
    def get_all_by_project(cls, context, project_id, marker, limit,
                           sort_keys=None, sort_dirs=None, filters=None,
                           offset=None):
        restores = db.restore_get_all_by_project(context, project_id, marker,
                                                 limit, sort_keys=sort_keys,
                                                 sort_dirs=sort_dirs,
                                                 filters=filters,
                                                 offset=offset)
        return base.obj_make_list(context, cls(context), objects.Restore,
                                  restores)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Handles all requests relating to OperationEngine."""


from karbor.db import base
from karbor.services.operationengine import rpcapi as oe_rpcapi


class API(base.Base):
    """API for interacting with the OperationEngine manager."""

    def __init__(self, db_driver=None):
        self.operationengine_rpcapi = oe_rpcapi.OperationEngineAPI()
        super(API, self).__init__(db_driver)

    def create_scheduled_operation(self, context, operation):
        self.operationengine_rpcapi.create_scheduled_operation(
            context, operation)

    def delete_scheduled_operation(self, context, operation_id, trigger_id):
        self.operationengine_rpcapi.delete_scheduled_operation(
            context, operation_id, trigger_id)

    def suspend_scheduled_operation(self, context, operation_id, trigger_id):
        self.operationengine_rpcapi.suspend_scheduled_operation(
            context, operation_id, trigger_id)

    def resume_scheduled_operation(self, context, operation_id, trigger_id):
        self.operationengine_rpcapi.resume_scheduled_operation(
            context, operation_id, trigger_id)

    def create_trigger(self, context, trigger):
        self.operationengine_rpcapi.create_trigger(context, trigger)

    def delete_trigger(self, context, trigger_id):
        self.operationengine_rpcapi.delete_trigger(context, trigger_id)

    def update_trigger(self, context, trigger):
        self.operationengine_rpcapi.update_trigger(context, trigger)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from functools import partial
import os

from io import BytesIO
from karbor.common import constants
from karbor import exception
from karbor.services.protection.client_factory import ClientFactory
from karbor.services.protection import protection_plugin
from karbor.services.protection.protection_plugins.image \
    import image_plugin_schemas as image_schemas
from karbor.services.protection.protection_plugins import utils
from oslo_config import cfg
from oslo_log import log as logging


image_backup_opts = [
    cfg.IntOpt('backup_image_object_size',
               default=65536*10,
               help='The size in bytes of instance image objects. '
                    'The value must be a multiple of 65536('
                    'the size of image\'s chunk).'),
    cfg.IntOpt('poll_interval', default=10,
               help='Poll interval for image status'),
]

LOG = logging.getLogger(__name__)


def get_image_status(glance_client, image_id):
    LOG.debug('Polling image (image_id: %s)', image_id)
    try:
        status = glance_client.images.get(image_id)
    except exception.NotFound:
        status = 'not-found'
    LOG.debug('Polled image (image_id: %s) status: %s',
              image_id, status)
    return status


class ProtectOperation(protection_plugin.Operation):
    def __init__(self, backup_image_object_size,
                 poll_interval):
        super(ProtectOperation, self).__init__()
        self._data_block_size_bytes = backup_image_object_size
        self._interval = poll_interval

    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        image_id = resource.id
        bank_section = checkpoint.get_resource_bank_section(image_id)

        resource_definition = {"resource_id": image_id}
        glance_client = ClientFactory.create_client('glance', context)
        LOG.info("Creating image backup, image_id: %s.", image_id)
        try:
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_PROTECTING)
            image_info = glance_client.images.get(image_id)
            if image_info.status != "active":
                is_success = utils.status_poll(
                    partial(get_image_status, glance_client, image_info.id),
                    interval=self._interval, success_statuses={'active'},
                    ignore_statuses={'queued', 'saving'},
                    failure_statuses={'killed', 'deleted', 'pending_delete',
                                      'deactivated', 'NotFound'}
                )
                if is_success is not True:
                    LOG.error("The status of image (id: %s) is invalid.",
                              image_id)
                    raise exception.CreateBackupFailed(
                        reason="The status of image is invalid.",
                        resource_id=image_id,
                        resource_type=constants.IMAGE_RESOURCE_TYPE)

            image_metadata = {
                "disk_format": image_info.disk_format,
                "container_format": image_info.container_format,
                "checksum": image_info.checksum
            }
            resource_definition["image_metadata"] = image_metadata

            bank_section.update_object("metadata", resource_definition)
        except Exception as err:
            LOG.error("Create image backup failed, image_id: %s.", image_id)
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.CreateBackupFailed(
                reason=err,
                resource_id=image_id,
                resource_type=constants.IMAGE_RESOURCE_TYPE)
        self._create_backup(glance_client, bank_section, image_id)

    def _create_backup(self, glance_client, bank_section, image_id):
        try:
            image_response = glance_client.images.data(image_id,
                                                       do_checksum=True)
            bank_chunk_num = int(self._data_block_size_bytes/65536)
            LOG.debug("Creating image backup, bank_chunk_num: %s.",
                      bank_chunk_num)

            # backup the data of image
            image_chunks_num = 0
            chunks_num = 1
            image_response_data = BytesIO()
            for chunk in image_response:
                image_response_data.write(chunk)
                image_chunks_num += 1
                if image_chunks_num == bank_chunk_num:
                    image_chunks_num = 0
                    image_response_data.seek(0, os.SEEK_SET)
                    data = image_response_data.read(
                        self._data_block_size_bytes)
                    bank_section.update_object("data_" + str(chunks_num), data)
                    image_response_data.truncate(0)
                    image_response_data.seek(0, os.SEEK_SET)
                    chunks_num += 1

            image_response_data.seek(0, os.SEEK_SET)
            data = image_response_data.read()
            if data != '':
                bank_section.update_object("data_" + str(chunks_num), data)

            # Save the chunks_num to metadata
            resource_definition = bank_section.get_object("metadata")
            if resource_definition is not None:
                resource_definition["chunks_num"] = chunks_num
            bank_section.update_object("metadata", resource_definition)

            # Update resource_definition backup_status
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_AVAILABLE)
            LOG.info('Protecting image (id: %s) to bank completed '
                     'successfully', image_id)
        except Exception as err:
            # update resource_definition backup_status
            LOG.exception('Protecting image (id: %s) to bank failed.',
                          image_id)
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.CreateBackupFailed(
                reason=err,
                resource_id=image_id,
                resource_type=constants.IMAGE_RESOURCE_TYPE)


class DeleteOperation(protection_plugin.Operation):
    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        image_id = resource.id
        bank_section = checkpoint.get_resource_bank_section(image_id)

        LOG.info("Deleting image backup failed, image_id: %s.", image_id)
        try:
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_DELETING)
            objects = bank_section.list_objects()
            for obj in objects:
                if obj == "status":
                    continue
                bank_section.delete_object(obj)
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_DELETED)
        except Exception as err:
            LOG.error("delete image backup failed, image_id: %s.", image_id)
            bank_section.update_object("status",
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.DeleteBackupFailed(
                reason=err,
                resource_id=image_id,
                resource_type=constants.IMAGE_RESOURCE_TYPE)


class RestoreOperation(protection_plugin.Operation):
    def __init__(self, poll_interval):
        super(RestoreOperation, self).__init__()
        self._interval = poll_interval

    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        original_image_id = resource.id
        heat_template = kwargs.get("heat_template")

        name = parameters.get("restore_name", "karbor-restore-image")
        LOG.info("Restoring image backup, image_id: %s.", original_image_id)

        glance_client = ClientFactory.create_client('glance', context)
        bank_section = checkpoint.get_resource_bank_section(original_image_id)
        try:
            resource_definition = bank_section.get_object('metadata')
            image_metadata = resource_definition['image_metadata']
            objects = [key.split("/")[-1] for key in
                       bank_section.list_objects()
                       if (key.split("/")[-1]).startswith("data_")]

            # check the chunks_num
            chunks_num = resource_definition.get("chunks_num", 0)
            if len(objects) != int(chunks_num):
                raise exception.RestoreBackupFailed(
                    reason=" The chunks_num of restored image is invalid.",
                    resource_id=original_image_id,
                    resource_type=constants.IMAGE_RESOURCE_TYPE)

            sorted_objects = sorted(objects, key=lambda s: int(s[5:]))
            image_data = ImageBankIO(bank_section, sorted_objects)
            disk_format = image_metadata["disk_format"]
            container_format = image_metadata["container_format"]
            image = glance_client.images.create(
                disk_format=disk_format,
                container_format=container_format,
                name=name)
            glance_client.images.upload(image.id, image_data)

            image_info = glance_client.images.get(image.id)
            if image_info.status != "active":
                is_success = utils.status_poll(
                    partial(get_image_status, glance_client, image_info.id),
                    interval=self._interval, success_statuses={'active'},
                    ignore_statuses={'queued', 'saving'},
                    failure_statuses={'killed', 'deleted', 'pending_delete',
                                      'deactivated', 'not-found'}
                )
                if is_success is not True:
                    LOG.error('The status of image is invalid. status:%s',
                              image_info.status)
                    raise exception.RestoreBackupFailed(
                        resource_id=image_info.id,
                        resource_type=constants.IMAGE_RESOURCE_TYPE)

            # check the checksum
            if image_info.checksum != image_metadata["checksum"]:
                raise exception.RestoreBackupFailed(
                    reason=" The checksum of restored image is invalid.",
                    resource_id=original_image_id,
                    resource_type=constants.IMAGE_RESOURCE_TYPE)
            heat_template.put_parameter(original_image_id, image.id)
        except Exception as e:
            LOG.error("Restore image backup failed, image_id: %s.",
                      original_image_id)
            raise exception.RestoreBackupFailed(
                reason=e, resource_id=original_image_id,
                resource_type=constants.IMAGE_RESOURCE_TYPE)
        LOG.info("Finish restoring image backup, image_id: %s.",
                 original_image_id)


class ImageBankIO(object):
        def __init__(self, bank_section, sorted_objects):
            super(ImageBankIO, self).__init__()
            self.bank_section = bank_section
            self.sorted_objects = sorted_objects
            self.obj_size = len(sorted_objects)
            self.length = 0

        def readable(self):
            return True

        def __iter__(self):
            return self

        def read(self, length=None):
            obj_index = self.length
            self.length += 1
            if self.length > self.obj_size:
                return ''
            return self.bank_section.get_object(
                self.sorted_objects[obj_index])


class GlanceProtectionPlugin(protection_plugin.ProtectionPlugin):
    _SUPPORT_RESOURCE_TYPES = [constants.IMAGE_RESOURCE_TYPE]

    def __init__(self, config=None):
        super(GlanceProtectionPlugin, self).__init__(config)
        self._config.register_opts(image_backup_opts,
                                   'image_backup_plugin')
        self._plugin_config = self._config.image_backup_plugin
        self._data_block_size_bytes = (
            self._plugin_config.backup_image_object_size)
        self._poll_interval = self._plugin_config.poll_interval

        if self._data_block_size_bytes % 65536 != 0 or (
                self._data_block_size_bytes <= 0):
            raise exception.InvalidParameterValue(
                err="The value of CONF.backup_image_object_size "
                    "is invalid!")

    @classmethod
    def get_supported_resources_types(cls):
        return cls._SUPPORT_RESOURCE_TYPES

    @classmethod
    def get_options_schema(cls, resources_type):
        return image_schemas.OPTIONS_SCHEMA

    @classmethod
    def get_restore_schema(cls, resources_type):
        return image_schemas.RESTORE_SCHEMA

    @classmethod
    def get_saved_info_schema(cls, resources_type):
        return image_schemas.SAVED_INFO_SCHEMA

    @classmethod
    def get_saved_info(cls, metadata_store, resource):
        pass

    def get_protect_operation(self, resource):
        return ProtectOperation(self._data_block_size_bytes,
                                self._poll_interval)

    def get_restore_operation(self, resource):
        return RestoreOperation(self._poll_interval)

    def get_delete_operation(self, resource):
        return DeleteOperation()



# Copyright (c) 2013 OpenStack Foundation
#
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
WSGI middleware for OpenStack API controllers.
"""

import routes


class APIMapper(routes.Mapper):
    def routematch(self, url=None, environ=None):
        if url is "":
            result = self._match("", environ)
            return result[0], result[1]
        return routes.Mapper.routematch(self, url, environ)

    def connect(self, *args, **kwargs):
        # NOTE(inhye): Default the format part of a route to only accept json
        #             and xml so it doesn't eat all characters after a '.'
        #             in the url.
        kwargs.setdefault('requirements', {})
        if not kwargs['requirements'].get('format'):
            kwargs['requirements']['format'] = 'json|xml'
        return routes.Mapper.connect(self, *args, **kwargs)


class ProjectMapper(APIMapper):
    def resource(self, member_name, collection_name, **kwargs):
        if 'parent_resource' not in kwargs:
            kwargs['path_prefix'] = '{project_id}/'
        else:
            parent_resource = kwargs['parent_resource']
            p_collection = parent_resource['collection_name']
            p_member = parent_resource['member_name']
            kwargs['path_prefix'] = '{project_id}/%s/:%s_id' % (p_collection,
                                                                p_member)
        routes.Mapper.resource(self,
                               member_name,
                               collection_name,
                               **kwargs)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from functools import partial
import six

from cinderclient import exceptions as cinder_exc
from oslo_config import cfg
from oslo_log import log as logging

from karbor.common import constants
from karbor import exception
from karbor.services.protection.client_factory import ClientFactory
from karbor.services.protection import protection_plugin
from karbor.services.protection.protection_plugins import utils
from karbor.services.protection.protection_plugins.volume \
    import volume_plugin_cinder_schemas as cinder_schemas


LOG = logging.getLogger(__name__)

cinder_backup_opts = [
    cfg.IntOpt(
        'poll_interval', default=15,
        help='Poll interval for Cinder backup status'
    ),
    cfg.BoolOpt(
        'backup_from_snapshot', default=True,
        help='First take a snapshot of the volume, and backup from '
        'it. Minimizes the time the volume is unavailable.'
    ),
]


def get_backup_status(cinder_client, backup_id):
    return get_resource_status(cinder_client.backups, backup_id, 'backup')


def get_volume_status(cinder_client, volume_id):
    return get_resource_status(cinder_client.volumes, volume_id, 'volume')


def get_snapshot_status(cinder_client, snapshot_id):
    return get_resource_status(cinder_client.volume_snapshots, snapshot_id,
                               'snapshot')


def get_resource_status(resource_manager, resource_id, resource_type):
    LOG.debug('Polling %(resource_type)s (id: %(resource_id)s)', {
        'resource_type': resource_type,
        'resource_id': resource_id,
    })
    try:
        resource = resource_manager.get(resource_id)
        status = resource.status
    except cinder_exc.NotFound:
        status = 'not-found'
    LOG.debug(
        'Polled %(resource_type)s (id: %(resource_id)s) status: %(status)s',
        {
            'resource_type': resource_type,
            'resource_id': resource_id,
            'status': status
        }
    )
    return status


class ProtectOperation(protection_plugin.Operation):
    def __init__(self, poll_interval, backup_from_snapshot):
        super(ProtectOperation, self).__init__()
        self._interval = poll_interval
        self._backup_from_snapshot = backup_from_snapshot
        self.snapshot_id = None

    def _create_snapshot(self, cinder_client, volume_id):
        snapshot = cinder_client.volume_snapshots.create(volume_id, force=True)

        snapshot_id = snapshot.id
        is_success = utils.status_poll(
            partial(get_snapshot_status, cinder_client, snapshot_id),
            interval=self._interval,
            success_statuses={'available', },
            failure_statuses={'error', 'error_deleting', 'deleting',
                              'not-found'},
            ignore_statuses={'creating', },
        )
        if not is_success:
            raise Exception

        return snapshot_id

    def _delete_snapshot(self, cinder_client, snapshot_id):
        LOG.info('Cleaning up snapshot (snapshot_id: %s)', snapshot_id)
        cinder_client.volume_snapshots.delete(snapshot_id)
        return utils.status_poll(
            partial(get_snapshot_status, cinder_client, snapshot_id),
            interval=self._interval,
            success_statuses={'not-found', },
            failure_statuses={'error', 'error_deleting', 'creating'},
            ignore_statuses={'deleting', },
        )

    def _create_backup(self, cinder_client, volume_id, backup_name,
                       description, snapshot_id=None, incremental=False,
                       container=None, force=False):
        backup = cinder_client.backups.create(
            volume_id=volume_id,
            name=backup_name,
            description=description,
            force=force,
            snapshot_id=snapshot_id,
            incremental=incremental,
            container=container
        )

        backup_id = backup.id
        is_success = utils.status_poll(
            partial(get_backup_status, cinder_client, backup_id),
            interval=self._interval,
            success_statuses={'available'},
            failure_statuses={'error'},
            ignore_statuses={'creating'},
        )

        if not is_success:
            try:
                backup = cinder_client.backups.get(backup_id)
            except Exception:
                reason = 'Unable to find backup'
            else:
                reason = backup.fail_reason
            raise Exception(reason)

        return backup_id

    def on_prepare_finish(self, checkpoint, resource, context, parameters,
                          **kwargs):
        volume_id = resource.id
        if not self._backup_from_snapshot:
            LOG.info('Skipping taking snapshot of volume %s - backing up '
                     'directly', volume_id)
            return

        LOG.info('Taking snapshot of volume %s', volume_id)
        bank_section = checkpoint.get_resource_bank_section(volume_id)
        bank_section.update_object('status',
                                   constants.RESOURCE_STATUS_PROTECTING)
        cinder_client = ClientFactory.create_client('cinder', context)
        try:
            self.snapshot_id = self._create_snapshot(cinder_client, volume_id)
        except Exception:
            bank_section.update_object('status',
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.CreateBackupFailed(
                reason='Error creating snapshot for volume',
                resource_id=volume_id,
                resource_type=constants.VOLUME_RESOURCE_TYPE,
            )

    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        volume_id = resource.id
        bank_section = checkpoint.get_resource_bank_section(volume_id)
        cinder_client = ClientFactory.create_client('cinder', context)
        LOG.info('creating volume backup, volume_id: %s', volume_id)
        bank_section.update_object('status',
                                   constants.RESOURCE_STATUS_PROTECTING)
        resource_metadata = {
            'volume_id': volume_id,
        }
        is_success = utils.status_poll(
            partial(get_volume_status, cinder_client, volume_id),
            interval=self._interval,
            success_statuses={'available', 'in-use', 'error_extending',
                              'error_restoring'},
            failure_statuses={'error', 'error_deleting', 'deleting',
                              'not-found'},
            ignore_statuses={'attaching', 'creating', 'backing-up',
                             'restoring-backup'},
        )
        if not is_success:
            bank_section.update_object('status',
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.CreateBackupFailed(
                reason='Volume is in erroneous state',
                resource_id=volume_id,
                resource_type=constants.VOLUME_RESOURCE_TYPE,
            )

        backup_name = parameters.get('backup_name', None)
        description = parameters.get('description', None)
        backup_mode = parameters.get('backup_mode', "full")
        container = parameters.get('container', None)
        force = parameters.get('force', False)
        incremental = False
        if backup_mode == "incremental":
            incremental = True
        elif backup_mode == "full":
            incremental = False

        try:
            backup_id = self._create_backup(cinder_client, volume_id,
                                            backup_name, description,
                                            self.snapshot_id,
                                            incremental, container, force)
        except Exception as e:
            LOG.error('Error creating backup (volume_id: %(volume_id)s '
                      'snapshot_id: %(snapshot_id)s): %(reason)s',
                      {'volume_id': volume_id,
                       'snapshot_id': self.snapshot_id,
                       'reason': e}
                      )
            bank_section.update_object('status',
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.CreateBackupFailed(
                reason=e,
                resource_id=volume_id,
                resource_type=constants.VOLUME_RESOURCE_TYPE,
            )

        resource_metadata['backup_id'] = backup_id
        bank_section.update_object('metadata', resource_metadata)
        bank_section.update_object('status',
                                   constants.RESOURCE_STATUS_AVAILABLE)
        LOG.info('Backed up volume (volume_id: %(volume_id)s snapshot_id: '
                 '%(snapshot_id)s backup_id: %(backup_id)s) successfully',
                 {'backup_id': backup_id,
                  'snapshot_id': self.snapshot_id,
                  'volume_id': volume_id}
                 )

        if self.snapshot_id:
            try:
                self._delete_snapshot(cinder_client, self.snapshot_id)
            except Exception as e:
                LOG.warning('Failed deleting snapshot: %(snapshot_id)s. '
                            'Reason: %(reason)s',
                            {'snapshot_id': self.snapshot_id, 'reason': e})


class RestoreOperation(protection_plugin.Operation):
    def __init__(self, poll_interval):
        super(RestoreOperation, self).__init__()
        self._interval = poll_interval

    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        resource_id = resource.id
        bank_section = checkpoint.get_resource_bank_section(resource_id)
        resource_metadata = bank_section.get_object('metadata')
        cinder_client = ClientFactory.create_client('cinder', context)

        # create volume
        volume_property = {
            'name': parameters.get(
                'restore_name',  '%s@%s' % (checkpoint.id, resource_id))
        }
        if 'restore_description' in parameters:
            volume_property['description'] = parameters['restore_description']
        backup_id = resource_metadata['backup_id']
        try:
            volume_id = cinder_client.restores.restore(backup_id).volume_id
            cinder_client.volumes.update(volume_id, **volume_property)
        except Exception as ex:
            LOG.error('Error creating volume (backup_id: %(backup_id)s): '
                      '%(reason)s',
                      {'backup_id': backup_id,
                       'reason': ex})
            raise

        # check and update status
        update_method = partial(
            utils.update_resource_restore_result,
            kwargs.get('restore'), resource.type, volume_id)

        update_method(constants.RESOURCE_STATUS_RESTORING)

        is_success = self._check_create_complete(cinder_client, volume_id)
        if is_success:
            update_method(constants.RESOURCE_STATUS_AVAILABLE)
            kwargs.get("heat_template").put_parameter(resource_id, volume_id)
        else:
            reason = 'Error creating volume'
            update_method(constants.RESOURCE_STATUS_ERROR, reason)

            raise exception.RestoreBackupFailed(
                reason=reason,
                resource_id=resource_id,
                resource_type=resource.type
            )

    def _check_create_complete(self, cinder_client, volume_id):
        return utils.status_poll(
            partial(get_volume_status, cinder_client, volume_id),
            interval=self._interval,
            success_statuses={'available'},
            failure_statuses={'error', 'not-found'},
            ignore_statuses={'creating', 'restoring-backup', 'downloading'},
        )


class DeleteOperation(protection_plugin.Operation):
    def __init__(self, poll_interval):
        super(DeleteOperation, self).__init__()
        self._interval = poll_interval

    def on_main(self, checkpoint, resource, context, parameters, **kwargs):
        resource_id = resource.id
        bank_section = checkpoint.get_resource_bank_section(resource_id)
        backup_id = None
        try:
            bank_section.update_object('status',
                                       constants.RESOURCE_STATUS_DELETING)
            resource_metadata = bank_section.get_object('metadata')
            backup_id = resource_metadata['backup_id']
            cinder_client = ClientFactory.create_client('cinder', context)
            try:
                backup = cinder_client.backups.get(backup_id)
                cinder_client.backups.delete(backup)
            except cinder_exc.NotFound:
                LOG.info('Backup id: %s not found. Assuming deleted',
                         backup_id)
            is_success = utils.status_poll(
                partial(get_backup_status, cinder_client, backup_id),
                interval=self._interval,
                success_statuses={'deleted', 'not-found'},
                failure_statuses={'error', 'error_deleting'},
                ignore_statuses={'deleting'},
            )
            if not is_success:
                raise exception.NotFound()
            bank_section.delete_object('metadata')
            bank_section.update_object('status',
                                       constants.RESOURCE_STATUS_DELETED)
        except Exception as e:
            LOG.error('delete volume backup failed, backup_id: %s', backup_id)
            bank_section.update_object('status',
                                       constants.RESOURCE_STATUS_ERROR)
            raise exception.DeleteBackupFailed(
                reason=six.text_type(e),
                resource_id=resource_id,
                resource_type=constants.VOLUME_RESOURCE_TYPE
            )


class CinderBackupProtectionPlugin(protection_plugin.ProtectionPlugin):
    _SUPPORT_RESOURCE_TYPES = [constants.VOLUME_RESOURCE_TYPE]

    def __init__(self, config=None):
        super(CinderBackupProtectionPlugin, self).__init__(config)
        self._config.register_opts(cinder_backup_opts,
                                   'cinder_backup_protection_plugin')
        self._plugin_config = self._config.cinder_backup_protection_plugin
        self._poll_interval = self._plugin_config.poll_interval
        self._backup_from_snapshot = self._plugin_config.backup_from_snapshot

    @classmethod
    def get_supported_resources_types(cls):
        return cls._SUPPORT_RESOURCE_TYPES

    @classmethod
    def get_options_schema(cls, resources_type):
        return cinder_schemas.OPTIONS_SCHEMA

    @classmethod
    def get_restore_schema(cls, resources_type):
        return cinder_schemas.RESTORE_SCHEMA

    @classmethod
    def get_saved_info_schema(cls, resources_type):
        return cinder_schemas.SAVED_INFO_SCHEMA

    @classmethod
    def get_saved_info(cls, metadata_store, resource):
        # TODO(hurong)
        pass

    def get_protect_operation(self, resource):
        return ProtectOperation(self._poll_interval,
                                self._backup_from_snapshot)

    def get_restore_operation(self, resource):
        return RestoreOperation(self._poll_interval)

    def get_delete_operation(self, resource):
        return DeleteOperation(self._poll_interval)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
OperationEngine Service
"""

from oslo_config import cfg
from oslo_log import log as logging
import oslo_messaging as messaging
from stevedore import driver as import_driver

from karbor.common import constants
from karbor import context as karbor_context
from karbor import exception
from karbor import manager
from karbor import objects
from karbor.services.operationengine.engine.triggers import trigger_manager
from karbor.services.operationengine import operation_manager
from karbor.services.operationengine import user_trust_manager


LOG = logging.getLogger(__name__)


trigger_manager_opts = [
    cfg.StrOpt('executor',
               default='green_thread',
               choices=['thread_pool', 'green_thread'],
               help='The name of executor which is used to run operations')
]

cfg.CONF.register_opts(trigger_manager_opts, 'operationengine')


class OperationEngineManager(manager.Manager):
    """karbor OperationEngine Manager."""

    RPC_API_VERSION = '1.0'

    target = messaging.Target(version=RPC_API_VERSION)

    def __init__(self, service_name=None,
                 *args, **kwargs):
        super(OperationEngineManager, self).__init__(*args, **kwargs)
        self._service_id = None
        self._trigger_manager = None
        self._user_trust_manager = None
        self._operation_manager = None
        self._executor = None

    @property
    def operation_manager(self):
        if not self._operation_manager:
            self._operation_manager = operation_manager.OperationManager(
                self.user_trust_manager
            )
        return self._operation_manager

    @property
    def executor(self):
        if not self._executor:
            executor_cls = import_driver.DriverManager(
                'karbor.operationengine.engine.executor',
                cfg.CONF.operationengine.executor).driver
            self._executor = executor_cls(self.operation_manager)
        return self._executor

    @property
    def user_trust_manager(self):
        if not self._user_trust_manager:
            self._user_trust_manager = user_trust_manager.UserTrustManager()
        return self._user_trust_manager

    @property
    def trigger_manager(self):
        if not self._trigger_manager:
            self._trigger_manager = trigger_manager.TriggerManager(
                self.executor
            )
        return self._trigger_manager

    def init_host(self, **kwargs):
        self._service_id = kwargs.get("service_id")
        self._restore()

    def cleanup_host(self):
        if self._trigger_manager:
            self._trigger_manager.shutdown()
            self._trigger_manager = None

    def _restore(self):
        self._restore_triggers()
        self._restore_operations()

    def _restore_triggers(self):
        limit = 100
        marker = None
        filters = {}
        ctxt = karbor_context.get_admin_context()
        while True:
            triggers = objects.TriggerList.get_by_filters(
                ctxt, filters, limit, marker)
            if not triggers:
                break

            for trigger in triggers:
                self.trigger_manager.add_trigger(trigger.id, trigger.type,
                                                 trigger.properties)
            if len(triggers) < limit:
                break
            marker = triggers[-1].id

    def _restore_operations(self):
        limit = 100
        marker = None
        filters = {"service_id": self._service_id,
                   "state": [constants.OPERATION_STATE_REGISTERED,
                             constants.OPERATION_STATE_RUNNING]}
        columns_to_join = ['operation']
        ctxt = karbor_context.get_admin_context()
        resume_states = [constants.OPERATION_STATE_RUNNING, ]
        while True:
            states = objects.ScheduledOperationStateList.get_by_filters(
                ctxt, filters, limit, marker, columns_to_join=columns_to_join)
            if not states:
                break

            for state in states:
                operation = state.operation
                if not operation.enabled:
                    continue

                resume = (state.state in resume_states)
                self.trigger_manager.register_operation(
                    operation.trigger_id, operation.id,
                    resume=resume, end_time_for_run=state.end_time_for_run)

                self.user_trust_manager.resume_operation(
                    operation.id, operation.user_id,
                    operation.project_id, state.trust_id)
            if len(states) < limit:
                break
            marker = states[-1].id

    @messaging.expected_exceptions(exception.TriggerNotFound,
                                   exception.InvalidInput,
                                   exception.TriggerIsInvalid,
                                   exception.AuthorizationFailure,
                                   exception.ScheduledOperationExist,
                                   exception.InvalidOperationDefinition)
    def create_scheduled_operation(self, context, operation):
        LOG.debug("Create scheduled operation.")
        self.operation_manager.check_operation_definition(
            operation.operation_type,
            operation.operation_definition,
        )

        # register operation
        self.trigger_manager.register_operation(operation.trigger_id,
                                                operation.id)
        trust_id = self.user_trust_manager.add_operation(
            context, operation.id)

        # create ScheduledOperationState record
        state_info = {
            "operation_id": operation.id,
            "service_id": self._service_id,
            "trust_id": trust_id,
            "state": constants.OPERATION_STATE_REGISTERED
        }
        operation_state = objects.ScheduledOperationState(
            context, **state_info)
        try:
            operation_state.create()
        except Exception:
            self.trigger_manager.unregister_operation(
                operation.trigger_id, operation.id)
            raise

    @messaging.expected_exceptions(exception.ScheduledOperationStateNotFound,
                                   exception.TriggerNotFound)
    def delete_scheduled_operation(self, context, operation_id, trigger_id):
        LOG.debug("Delete scheduled operation.")

        operation_state = objects.ScheduledOperationState.get_by_operation_id(
            context, operation_id)
        if constants.OPERATION_STATE_DELETED != operation_state.state:
            operation_state.state = constants.OPERATION_STATE_DELETED
            operation_state.save()

        self.trigger_manager.unregister_operation(trigger_id, operation_id)
        self.user_trust_manager.delete_operation(context, operation_id)

    @messaging.expected_exceptions(exception.TriggerNotFound)
    def suspend_scheduled_operation(self, context, operation_id, trigger_id):
        LOG.debug("Suspend scheduled operation.")
        self.trigger_manager.unregister_operation(trigger_id, operation_id)

    @messaging.expected_exceptions(exception.TriggerNotFound,
                                   exception.TriggerIsInvalid)
    def resume_scheduled_operation(self, context, operation_id, trigger_id):
        LOG.debug("Resume scheduled operation.")

        try:
            self.trigger_manager.register_operation(
                trigger_id, operation_id)
        except exception.ScheduledOperationExist:
            pass
        except Exception:
            raise

    @messaging.expected_exceptions(exception.InvalidInput)
    def create_trigger(self, context, trigger):
        self.trigger_manager.add_trigger(trigger.id, trigger.type,
                                         trigger.properties)

    @messaging.expected_exceptions(exception.TriggerNotFound,
                                   exception.DeleteTriggerNotAllowed)
    def delete_trigger(self, context, trigger_id):
        self.trigger_manager.remove_trigger(trigger_id)

    @messaging.expected_exceptions(exception.TriggerNotFound,
                                   exception.InvalidInput)
    def update_trigger(self, context, trigger):
        self.trigger_manager.update_trigger(trigger.id, trigger.properties)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import session as keystone_session
from oslo_config import cfg

from karbor.context import RequestContext
from karbor.services.protection.clients import nova
from karbor.tests import base


class NovaClientTest(base.TestCase):
    def setUp(self):
        super(NovaClientTest, self).setUp()
        service_catalog = [
            {'type': 'compute',
             'name': 'nova',
             'endpoints': [{'publicURL': 'http://127.0.0.1:8774/v2.1/abcd'}],
             },
        ]
        self._context = RequestContext(user_id='demo',
                                       project_id='abcd',
                                       auth_token='efgh',
                                       service_catalog=service_catalog)

    def test_create_client_by_endpoint(self):
        cfg.CONF.set_default('nova_endpoint',
                             'http://127.0.0.1:8774/v2.1',
                             'nova_client')

        client = nova.create(self._context, cfg.CONF,
                             session=keystone_session.Session(auth=None))
        self.assertEqual('compute', client.client.service_type)
        self.assertEqual('http://127.0.0.1:8774/v2.1/abcd',
                         client.client.endpoint_override)

    def test_create_client_by_catalog(self):
        client = nova.create(self._context, cfg.CONF,
                             session=keystone_session.Session(auth=None))
        self.assertEqual('compute', client.client.service_type)
        self.assertEqual('http://127.0.0.1:8774/v2.1/abcd',
                         client.client.endpoint_override)



#    Copyright 2015 SimpliVity Corp.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from karbor import objects
from karbor.tests.unit import fake_operation_log
from karbor.tests.unit import objects as test_objects


class TestOperationLog(test_objects.BaseObjectsTestCase):
    @staticmethod
    def _compare(test, db, obj):
        db = {k: v for k, v in db.items()}
        test_objects.BaseObjectsTestCase._compare(test, db, obj)

    @mock.patch('karbor.objects.OperationLog.get_by_id')
    def test_get_by_id(self, operation_log_get):
        db_operation_log = fake_operation_log.fake_db_operation_log()
        operation_log_get.return_value = db_operation_log
        operation_log = objects.OperationLog.get_by_id(self.context, "1")
        operation_log_get.assert_called_once_with(self.context, "1")
        self._compare(self, db_operation_log, operation_log)

    @mock.patch('karbor.db.sqlalchemy.api.operation_log_create')
    def test_create(self, operation_log_create):
        db_operation_log = fake_operation_log.fake_db_operation_log()
        operation_log_create.return_value = db_operation_log
        operation_log = objects.OperationLog(context=self.context)
        operation_log.create()
        self.assertEqual(db_operation_log['id'], operation_log.id)

    @mock.patch('karbor.db.sqlalchemy.api.operation_log_update')
    def test_save(self, operation_log_update):
        db_operation_log = fake_operation_log.fake_db_operation_log()
        operation_log = objects.OperationLog._from_db_object(
            self.context, objects.OperationLog(), db_operation_log)
        operation_log.state = 'finished'
        operation_log.save()
        operation_log_update.assert_called_once_with(
            self.context, operation_log.id, {'state': 'finished'})

    @mock.patch('karbor.db.sqlalchemy.api.operation_log_destroy')
    def test_destroy(self, operation_log_destroy):
        db_operation_log = fake_operation_log.fake_db_operation_log()
        operation_log = objects.OperationLog._from_db_object(
            self.context, objects.OperationLog(), db_operation_log)
        operation_log.destroy()
        self.assertTrue(operation_log_destroy.called)
        admin_context = operation_log_destroy.call_args[0][0]
        self.assertTrue(admin_context.is_admin)

    def test_obj_field_status(self):
        operation_log = objects.OperationLog(context=self.context,
                                             state='finished')
        self.assertEqual('finished', operation_log.state)



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.


def register_all():
    # You must make sure your object gets imported in this
    # function in order for it to be registered by services that may
    # need to receive it via RPC.
    __import__('karbor.objects.service')
    __import__('karbor.objects.plan')
    __import__('karbor.objects.scheduled_operation')
    __import__('karbor.objects.trigger')
    __import__('karbor.objects.scheduled_operation_log')
    __import__('karbor.objects.scheduled_operation_state')
    __import__('karbor.objects.restore')
    __import__('karbor.objects.operation_log')
    __import__('karbor.objects.checkpoint_record')



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from heatclient import client as hc
from keystoneclient.v3 import client as kc_v3
from oslo_config import cfg
from oslo_log import log as logging

from karbor.common import config
from karbor.services.protection.clients import utils

LOG = logging.getLogger(__name__)

SERVICE = 'heat'
heat_client_opts = [
    cfg.StrOpt('region_id',
               default='RegionOne',
               help='The region id which the service belongs to.'),
    cfg.StrOpt(SERVICE + '_endpoint',
               help='URL of the heat endpoint.'),
    cfg.StrOpt(SERVICE + '_catalog_info',
               default='orchestration:heat:publicURL',
               help='Info to match when looking for heat in the service '
               'catalog. Format is: separated values of the form: '
               '<service_type>:<service_name>:<endpoint_type> - '
               'Only used if cinder_endpoint is unset'),
    cfg.StrOpt(SERVICE + '_ca_cert_file',
               help='Location of the CA certificate file '
                    'to use for client requests in SSL connections.'),
    cfg.BoolOpt(SERVICE + '_auth_insecure',
                default=False,
                help='Bypass verification of server certificate when '
                     'making SSL connection to Cinder.'),
]

CONFIG_GROUP = '%s_client' % SERVICE
CONF = cfg.CONF
CONF.register_opts(config.service_client_opts, group=CONFIG_GROUP)
CONF.register_opts(heat_client_opts, group=CONFIG_GROUP)
CONF.set_default('service_name', 'heat', CONFIG_GROUP)
CONF.set_default('service_type', 'orchestration', CONFIG_GROUP)

KEYSTONECLIENT_VERSION = (3, 0)
HEATCLIENT_VERSION = '1'


def create(context, conf, **kwargs):
    cfg.CONF.register_opts(heat_client_opts, group=CONFIG_GROUP)

    client_config = conf[CONFIG_GROUP]
    if kwargs.get("auth_url", None):
        return _create_client_with_auth_url(context, client_config, **kwargs)
    else:
        return _create_client_with_tenant(context, client_config, **kwargs)


def _create_client_with_auth_url(context, client_config, **kwargs):
    auth_url = kwargs["auth_url"]
    username = kwargs["username"]
    password = kwargs["password"]
    tenant_name = context.project_name
    cacert = client_config.heat_ca_cert_file
    insecure = client_config.heat_auth_insecure
    LOG.debug('Creating heat client with auth url %s.', auth_url)
    try:
        keystone = kc_v3.Client(version=KEYSTONECLIENT_VERSION,
                                username=username,
                                tenant_name=tenant_name,
                                password=password,
                                auth_url=auth_url)

        auth_token = keystone.auth_token
        heat_endpoint = ''
        services = keystone.service_catalog.catalog['catalog']
        for service in services:
            if service['name'] == 'heat':
                endpoints = service['endpoints']
                for endpoint in endpoints:
                    if endpoint['interface'] == 'public':
                        heat_endpoint = endpoint['url']
        heat = hc.Client(HEATCLIENT_VERSION, endpoint=heat_endpoint,
                         token=auth_token, cacert=cacert, insecure=insecure)
        return heat
    except Exception:
        LOG.error('Creating heat client with url %s.', auth_url)
        raise


def _create_client_with_tenant(context, client_config, **kwargs):
    url = utils.get_url(SERVICE, context, client_config,
                        append_project_fmt='%(url)s/%(project)s', **kwargs)
    LOG.debug('Creating heat client with url %s.', url)

    if kwargs.get('session'):
        return hc.Client(HEATCLIENT_VERSION, session=kwargs.get('session'),
                         endpoint=url)

    args = {
        'endpoint': url,
        'token': context.auth_token,
        'cacert': client_config.heat_ca_cert_file,
        'insecure': client_config.heat_auth_insecure,
    }
    return hc.Client(HEATCLIENT_VERSION, **args)



# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_service import loopingcall


LOG = logging.getLogger(__name__)


def update_resource_restore_result(restore_record, resource_type, resource_id,
                                   status, reason=''):
    try:
        restore_record.update_resource_status(resource_type, resource_id,
                                              status, reason)
        restore_record.save()
    except Exception:
        LOG.error('Unable to update restoration result. '
                  'resource type: %(resource_type)s, '
                  'resource id: %(resource_id)s, '
                  'status: %(status)s, reason: %(reason)s',
                  {'resource_type': resource_type, 'resource_id': resource_id,
                   'status': status, 'reason': reason})
        pass


def status_poll(get_status_func, interval, success_statuses=set(),
                failure_statuses=set(), ignore_statuses=set(),
                ignore_unexpected=False):
    def _poll():
        status = get_status_func()
        if status in success_statuses:
            raise loopingcall.LoopingCallDone(retvalue=True)
        if status in failure_statuses:
            raise loopingcall.LoopingCallDone(retvalue=False)
        if status in ignore_statuses:
            return
        if ignore_unexpected is False:
            raise loopingcall.LoopingCallDone(retvalue=False)

    loop = loopingcall.FixedIntervalLoopingCall(_poll)
    return loop.start(interval=interval, initial_delay=interval).wait()



# Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from datetime import datetime
import mock

from oslo_utils import timeutils

from karbor.services.protection.bank_plugin import Bank
from karbor.services.protection.checkpoint import CheckpointCollection
from karbor.tests import base
from karbor.tests.unit.protection.fakes import fake_protection_plan
from karbor.tests.unit.protection.test_bank import _InMemoryBankPlugin
from karbor.tests.unit.protection.test_bank import _InMemoryLeasePlugin


class CheckpointCollectionTest(base.TestCase):
    def _create_test_collection(self):
        return CheckpointCollection(Bank(_InMemoryBankPlugin()),
                                    _InMemoryLeasePlugin())

    def test_create_checkpoint(self):
        collection = self._create_test_collection()
        checkpoint = collection.create(fake_protection_plan())
        checkpoint.status = "finished"
        checkpoint.commit()
        self.assertEqual(
            checkpoint.status,
            collection.get(checkpoint_id=checkpoint.id).status,
        )

    def test_list_checkpoints(self):
        collection = self._create_test_collection()
        plan = fake_protection_plan()
        provider_id = plan['provider_id']
        result = {collection.create(plan).id for i in range(10)}
        self.assertEqual(set(collection.list_ids(provider_id)), result)

    def test_list_checkpoints_by_plan_id(self):
        collection = self._create_test_collection()
        plan_1 = fake_protection_plan()
        plan_1["id"] = "fake_plan_id_1"
        plan_1['provider_id'] = "fake_provider_id_1"
        provider_id_1 = plan_1['provider_id']
        checkpoints_plan_1 = {collection.create(plan_1).id for i in range(10)}

        plan_2 = fake_protection_plan()
        plan_2["id"] = "fake_plan_id_2"
        plan_2['provider_id'] = "fake_provider_id_2"
        provider_id_2 = plan_1['provider_id']
        checkpoints_plan_2 = {collection.create(plan_2).id for i in range(10)}
        self.assertEqual(set(collection.list_ids(provider_id=provider_id_1,
                                                 plan_id="fake_plan_id_1")),
                         checkpoints_plan_1)
        self.assertEqual(set(collection.list_ids(provider_id=provider_id_2,
                                                 plan_id="fake_plan_id_2")),
                         checkpoints_plan_2)

    def test_list_checkpoints_by_date(self):
        collection = self._create_test_collection()
        date1 = datetime.strptime("2016-06-12", "%Y-%m-%d")
        timeutils.utcnow = mock.MagicMock()
        timeutils.utcnow.return_value = date1
        plan = fake_protection_plan()
        provider_id = plan['provider_id']
        checkpoints_date_1 = {collection.create(plan).id for i in range(10)}
        date2 = datetime.strptime("2016-06-13", "%Y-%m-%d")
        timeutils.utcnow = mock.MagicMock()
        timeutils.utcnow.return_value = date2
        checkpoints_date_2 = {collection.create(plan).id for i in range(10)}
        self.assertEqual(set(collection.list_ids(provider_id=provider_id,
                                                 start_date=date1,
                                                 end_date=date1)),
                         checkpoints_date_1)
        self.assertEqual(set(collection.list_ids(provider_id=provider_id,
                                                 start_date=date2,
                                                 end_date=date2)),
                         checkpoints_date_2)

    def test_delete_checkpoint(self):
        collection = self._create_test_collection()
        plan = fake_protection_plan()
        provider_id = plan['provider_id']
        result = {collection.create(plan).id for i in range(10)}
        checkpoint = collection.get(result.pop())
        checkpoint.purge()
        self.assertEqual(set(collection.list_ids(provider_id)), result)

    def test_write_checkpoint_with_invalid_lease(self):
        collection = self._create_test_collection()
        checkpoint = collection.create(fake_protection_plan())
        collection._bank_lease.check_lease_validity = mock.MagicMock()
        collection._bank_lease.check_lease_validity.return_value = False
        checkpoint.status = "finished"
        self.assertNotEqual(
            checkpoint.status,
            collection.get(checkpoint_id=checkpoint.id).status,
        )



#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import six

from karbor.common import constants
from karbor import exception
from karbor import resource
from karbor.services.protection.client_factory import ClientFactory
from karbor.services.protection import protectable_plugin
from oslo_log import log as logging

LOG = logging.getLogger(__name__)


class NetworkProtectablePlugin(protectable_plugin.ProtectablePlugin):
    """Protectable plugin implementation for Network from Neutron.

    """

    _SUPPORT_RESOURCE_TYPE = constants.NETWORK_RESOURCE_TYPE

    def _neutron_client(self, cntxt):
        return ClientFactory.create_client('neutron', cntxt)

    def _nova_client(self, cntxt):
        return ClientFactory.create_client('nova', cntxt)

    def get_resource_type(self):
        return self._SUPPORT_RESOURCE_TYPE

    def get_parent_resource_types(self):
        return (constants.SERVER_RESOURCE_TYPE,
                constants.PROJECT_RESOURCE_TYPE)

    def _get_network_id(self):
        """Set network_id as project_id

        Cause the network plugin include the ports, networks,
        subnets, routes, securitygroups, So make the id for
        the whole plugin-package info, not just like the server
        plugin which have the real server-id for the server.
        """

        network_id = self._context.project_id
        return network_id

    def list_resources(self, context, parameters=None):
        try:
            netclient = self._neutron_client(context)
            networks = netclient.list_networks().get('networks')
        except Exception as e:
            LOG.exception("List all summary networks from neutron failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        else:
            if networks:
                return [resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                          id=self._get_network_id(),
                                          name="Network Topology")]
            return []

    def show_resource(self, context, resource_id, parameters=None):
        try:
            if resource_id != self._get_network_id():
                return None

            netclient = self._neutron_client(context)
            networks = netclient.list_networks().get('networks')
        except Exception as e:
            LOG.exception("List all summary networks from neutron failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))
        else:
            if networks:
                return resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                         id=self._get_network_id(),
                                         name="Network Topology")
            return None

    def _get_dependent_resources_by_server(self,
                                           context,
                                           parent_resource):
        try:
            # get metadata about network from neutron
            net_client = self._neutron_client(context)
            network_infos = net_client.list_networks().get('networks')
            neutron_networks = {network["id"] for network in network_infos}

            # get interface info from server
            nova_networks = set()
            serverid = parent_resource.id
            nova_client = ClientFactory.create_client("nova", context)
            interface_list = nova_client.servers.interface_list(serverid)

            # check net_id in interface
            for iface in interface_list:
                net_id = iface.net_id
                if net_id not in nova_networks:
                    nova_networks.add(net_id)

            # get the exsited networks
            valid_networks = nova_networks.intersection(neutron_networks)
            if valid_networks:
                return [resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                          id=self._get_network_id(),
                                          name="Network Topology")]
            return []
        except Exception as e:
            LOG.exception("List all interfaces from nova failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))

    def _get_dependent_resources_by_project(self,
                                            context,
                                            parent_resource):
        try:
            tid = parent_resource.id
            netclient = self._neutron_client(context)
            networks = netclient.list_networks(tenant_id=tid).get('networks')

            if networks:
                return [resource.Resource(type=self._SUPPORT_RESOURCE_TYPE,
                                          id=self._get_network_id(),
                                          name="Network Topology")]
            else:
                return []

        except Exception as e:
            LOG.exception("List all summary networks from neutron failed.")
            raise exception.ListProtectableResourceFailed(
                type=self._SUPPORT_RESOURCE_TYPE,
                reason=six.text_type(e))

    def get_dependent_resources(self, context, parent_resource):
        if parent_resource.type == constants.SERVER_RESOURCE_TYPE:
            return self._get_dependent_resources_by_server(context,
                                                           parent_resource)

        if parent_resource.type == constants.PROJECT_RESOURCE_TYPE:
            return self._get_dependent_resources_by_project(context,
                                                            parent_resource)

        return []



#    Copyright (c) 2016 Shanghai EISOO Information Technology Corp.
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock

from karbor.context import RequestContext
from karbor.services.protection.clients import eisoo

from karbor.tests import base


class FakeConfig(object):
    def __init__(self):
        super(FakeConfig, self).__init__()
        self.eisoo_client = EisooClient()

    def __call__(self, args):
        pass

    def register_opts(self, opts, **kwargs):
        pass


class EisooClient(object):
    def __init__(self):
        super(EisooClient, self).__init__()
        self.eisoo_endpoint = 'eisoo_endpoint'
        self.eisoo_app_id = 'eisoo_app_id'
        self.eisoo_app_secret = 'eisoo_app_secret'


class ABClientTest(base.TestCase):
    def setUp(self):
        super(ABClientTest, self).setUp()
        self._context = RequestContext(user_id='demo',
                                       project_id='asdf',
                                       auth_token='qwe',
                                       service_catalog=None)

    @mock.patch('oslo_config.cfg.ConfigOpts', FakeConfig)
    @mock.patch('karbor.utils.find_config')
    @mock.patch('os.path.abspath')
    def test_create_client_by_config_file(self, mock_findconfig, mock_abspath):
        mock_findconfig.return_value = '/etc/provider.d'
        mock_abspath.return_value = ''

        client = eisoo.create(self._context, None)
        self.assertEqual('eisoo_app_id', client._app_id)

    def tearDown(self):
        super(ABClientTest, self).tearDown()
